Module DISPLACEMENT 

! --------------------------------------------------------------------------
! MPI Version of the Cartesian Solar Wave Simulator.
! Copyright 2008, Shravan Hanasoge
                                                                                                                                                         
! W. W. Hansen Experimental Physics Laboratory
! Stanford University, Stanford, CA 94305, USA
! Email: shravan@solar.stanford.edu
! --------------------------------------------------------------------------

! SUBROUTINES IN THIS MODULE:
!

  use initialize
  use derivatives
  use mp_physics
  use damping
  implicit none

Contains

!============================================================================


SUBROUTINE MP_QUIET_SPONGE_3D_DISPL

 ! CONSERVATIVE FORM OF THE FLUXES.

 ! I can't really do the conservative form - doesn't work for the incredibly small density 
 ! and forcing functions anyway. I don't know what the error is anyway
 implicit none
 integer k,i,j

 call ddxyz(xi_x, dxixdx, xi_y, dxiydy, xi_z, dxizdz)
 div = dxixdx + dxiydy + dxizdz

 p = - c2rho0 * div - xi_z * gradp0_z
 rho = - rho0 * div - xi_z * gradrho0_z 

 call ddxyz(p, gradp_x, p, gradp_y, p, gradp_z)

 RHSv_x = - rhoinv * gradp_x - spongexyz * v_x

 RHSv_y = - rhoinv * gradp_y - spongexyz * v_y

 RHSv_z = - rhoinv * gradp_z - spongexyz * v_z

 do k= 1,nz
   RHSv_z(:,:,k) = RHSv_z(:,:,k) - rho(:,:,k)*g(k)*rhoinv(:,:,k)
   if (ABS(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) +  forcing*source_dep(k)
 enddo

 RHSxi_x = v_x
 RHSxi_y = v_y
 RHSxi_z = v_z


 do k=1,6
  scr(:,:,nz,k) = 0.0
  scr(:,:,1,k) = 0.0
 enddo


 if (DAMP_WAVES) call damp_velocity

 END SUBROUTINE MP_QUIET_SPONGE_3D_DISPL

!================================================================================================

SUBROUTINE MP_MHD_SPONGE_3D_DISPL

 ! CONSERVATIVE FORM OF THE FLUXES.

 ! I can't really do the conservative form - doesn't work for the incredibly small density 
 ! and forcing functions anyway. I don't know what the error is anyway
 implicit none
 integer i,j,k
 real*8, dimension(nx, dim2(rank), nz) :: temp1, temp2, temp3, flux1, flux2, flux3

 call ddxyz(xi_x, dxixdx, xi_y, dxiydy, xi_z, dxizdz)
 div = dxixdx + dxiydy + dxizdz

 p = - c2rho0 * div - xi_z * gradp0_z - xi_x * gradp0_x - xi_y * gradp0_y
 rho = - rho0 * div - xi_z * gradrho0_z - xi_x * gradrho0_x - xi_y * gradrho0_y

 flux1 = - (xi_z * boy - xi_y * boz)
 flux2 = - (xi_x * boz - xi_z * box)
 flux3 = - (xi_y * box - xi_x * boy)

! call ddz(v_z, dvzdz, 1)
! call ddz(p, gradp_z, 1)

! call ddz(flux2, dzflux2, 1)
! call ddz(flux1, dzflux1, 1)

! call ddz(bx, dzbx, 1)
! call ddz(by, dzby, 1)


 ! CALLING OVERLAP DERIVATIVE ROUTINE
 
 ! NOTE THAT THE ORDER IS SENSITIVE.
 ! MUST FULLY COMPUTE bx, by, bz BEFORe
 ! COMPUTING DERIVATIVES THEREOF

 call ddxyz(flux3, by, p, gradp_y, flux2, dzflux2)
 call ddxyz(p, gradp_x, flux3, bx, flux1, dzflux1)
 call ddxyz(flux2, bz, flux1, flux3, p, gradp_z)

 bx =   bx - dzflux2
 by = - by + dzflux1
 bz =   bz - flux3

 ! NOTE THAT flux3 is overwritten in the third ddxyz call

 call ddxyz(by, curlbz, bz, curlbx,  bx, dzbx)
 call ddxyz(bz, curlby, bx, flux1, by, dzby)


 
! call ddx(flux3, RHSb_y, 1)
! call ddy(flux3, RHSb_x, 1)

! call ddx(flux2, RHSb_z,1)
! call ddy(flux1, flux2, 1)


! call ddx(v_x, dvxdx, 1)
! call ddx(p, gradp_x, 1)

! call ddy(v_y, dvydy, 1)
! call ddy(p, gradp_y, 1)

! call ddy(bz, curlbx, 1)
 curlbx = curlbx - dzby

! call ddx(bz, curlby, 1)
 curlby = -curlby + dzbx

! call ddx(by, curlbz, 1)
! call ddy(bx, flux1, 1)
 curlbz = curlbz - flux1


 call cross(curlbx, curlby, curlbz, box, boy, boz, flux1, flux2, flux3)

 RHSv_x = rhoinv * (- gradp_x + flux1)
 RHSv_y = rhoinv * (- gradp_y + flux2)
 RHSv_z = rhoinv * (- gradp_z + flux3)


 call cross(curlbox, curlboy, curlboz, bx, by, bz, flux1, flux2, flux3)


 RHSv_x = RHSv_x + rhoinv*flux1 - spongexyz * v_x
 RHSv_y = RHSv_y + rhoinv*flux2 - spongexyz * v_y
 RHSv_z = RHSv_z + rhoinv*flux3 - spongexyz * v_z


 do k= 1,nz
   RHSv_z(:,:,k) = RHSv_z(:,:,k) - rhoinv(:,:,k) * rho(:,:,k)*g(k)

   if (abs(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) + forcing*source_dep(k) 

 enddo 

 RHSxi_x = v_x
 RHSxi_y = v_y
 RHSxi_z = v_z

 do k=1,6
  scr(:,:,nz,k) = 0.
  scr(:,:,1,k) = 0.
 enddo


 if (DAMP_WAVES) call damp_velocity

 END SUBROUTINE MP_MHD_SPONGE_3D_DISPL


!================================================================================================


 Subroutine MP_FLOWS_SPONGE_3D_DISPL ()

  ! Right hand side computation

  implicit none
  integer i,j,k
  real*8, allocatable, dimension(:,:,:) :: temp_x, temp_y, temp_z

  allocate(temp_x(nx,dim2(rank),nz),temp_y(nx,dim2(rank),nz),&
   temp_z(nx,dim2(rank),nz))


  ! ONLY THE LINEAR ADVECTION TERM IS ACCOUNTED FOR. MORE TERMS
  ! NEED TO BE ADDED TO ENSURE CONSISTENCY BUT FOR ALL PRACTICAL
  ! PURPOSES, I DEEM THIS APPROXIMATION SUFFICIENT

  call ddxyz(xi_x, dxixdx, xi_y, dxiydy, xi_z, dxizdz)
  div = dxixdx + dxiydy + dxizdz

  p = - c2rho0 * div - xi_z * gradp0_z
  rho = - rho0 * div - xi_z * gradrho0_z 

 call ddxyz(p, gradp_x, p, gradp_y, p, gradp_z)

 call ddxyz(v_x, temp_x, v_x, temp_y, v_x, temp_z)

 RHSv_x = - 2.0 * (v0_x * temp_x + v0_y * temp_y + v0_z * temp_z) - &
		rhoinv * gradp_x - spongexyz * v_x


 call ddxyz(v_y, temp_x, v_y, temp_y, v_y, temp_z)

 RHSv_y = - 2.0 * (v0_x * temp_x + v0_y * temp_y + v0_z * temp_z) - &
		rhoinv * gradp_y - spongexyz * v_y


 call ddxyz(v_z, temp_x, v_z, temp_y, v_z, temp_z)

 RHSv_z = - 2.0 * (v0_x * temp_x + v0_y * temp_y + v0_z * temp_z) - &
		rhoinv * gradp_z - spongexyz * v_z


 deallocate(temp_x, temp_y, temp_z)

 do k = 1,nz

     RHSv_z(:,:,k)  = RHSv_z(:,:,k) - rhoinv(:,:,k) *  rho(:,:,k) * g(k)

     if (ABS(k-e_rad) .LE. 4) then
        RHSv_z(:,:,k) = RHSv_z(:,:,k) +  forcing*source_dep(k)
     endif

  enddo
  
  RHSxi_x = v_x
  RHSxi_y = v_y
  RHSxi_z = v_z

 do k=1,6
  scr(:,:,nz,k) = 0.
  scr(:,:,1,k) = 0.
 enddo


 if (DAMP_WAVES) call damp_velocity

 end SUBROUTINE MP_FLOWS_SPONGE_3D_DISPL

!================================================================================================

 Subroutine MP_FLOWS_PML_3D_DISPL ()

  ! Right hand side computation

  implicit none
  integer i,j,k,pmlindex
  real*8, allocatable, dimension(:,:,:) :: temp_x, temp_y, temp_z

  allocate(temp_x(nx,dim2(rank),nz),temp_y(nx,dim2(rank),nz),&
   temp_z(nx,dim2(rank),nz))


  ! ONLY THE LINEAR ADVECTION TERM IS ACCOUNTED FOR. MORE TERMS
  ! NEED TO BE ADDED TO ENSURE CONSISTENCY BUT FOR ALL PRACTICAL
  ! PURPOSES, I DEEM THIS APPROXIMATION SUFFICIENT

  ! NOTE THAT I ASSUME THAT v0 = 0.0 in the PML ZONE!


 call ddxyz(xi_x, dxixdx, xi_y, dxiydy, xi_z, dxizdz)

 pmlindex = 0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsivz(:,:,pmlindex) = az(:,:,k)*dxizdz(:,:,k) + bzpml(:,:,k)*psivz(:,:,pmlindex)    
   dxizdz(:,:,k) = dxizdz(:,:,k) + psivz(:,:,pmlindex)
  endif 

 enddo

 div = dxixdx + dxiydy + dxizdz

 p = - c2rho0 * div - xi_z * gradp0_z
 rho = - rho0 * div - xi_z * gradrho0_z 

 call ddxyz(p, gradp_x, p, gradp_y, p, gradp_z)

 pmlindex = 0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsip(:,:,pmlindex) = az(:,:,k)*gradp_z(:,:,k) + bzpml(:,:,k)*psip(:,:,pmlindex)    
   gradp_z(:,:,k) =  gradp_z(:,:,k) + psip(:,:,pmlindex)
  endif 

 enddo
 
 call ddxyz(v_x, temp_x, v_x, temp_y, v_x, temp_z)

 RHSv_x = - 2.0 * (v0_x * temp_x + v0_y * temp_y + v0_z * temp_z) - &
		rhoinv * gradp_x 


 call ddxyz(v_y, temp_x, v_y, temp_y, v_y, temp_z)

 RHSv_y = - 2.0 * (v0_x * temp_x + v0_y * temp_y + v0_z * temp_z) - &
		rhoinv * gradp_y 


 call ddxyz(v_z, temp_x, v_z, temp_y, v_z, temp_z)

 RHSv_z = - 2.0 * (v0_x * temp_x + v0_y * temp_y + v0_z * temp_z) - &
		rhoinv * gradp_z 

 deallocate(temp_x, temp_y, temp_z)

 do k = 1,nz

     RHSv_z(:,:,k)  = RHSv_z(:,:,k) - rhoinv(:,:,k) *  rho(:,:,k) * g(k)

     if (ABS(k-e_rad) .LE. 4) then
        RHSv_z(:,:,k) = RHSv_z(:,:,k) +  forcing*source_dep(k)
     endif

  enddo
  
  RHSxi_x = v_x
  RHSxi_y = v_y
  RHSxi_z = v_z

 do k=1,6
  scr(:,:,nz,k) = 0.
  scr(:,:,1,k) = 0.
 enddo


 if (DAMP_WAVES) call damp_velocity

 end SUBROUTINE MP_FLOWS_PML_3D_DISPL

!================================================================================================


SUBROUTINE MP_QUIET_PML_3D_DISPL

 ! CONSERVATIVE FORM OF THE FLUXES.

 ! I can't really do the conservative form - doesn't work for the incredibly small density 
 ! and forcing functions anyway. I don't know what the error is anyway
 implicit none
 integer k,i,j,pmlindex

 call ddxyz(xi_x, dxixdx, xi_y, dxiydy, xi_z, dxizdz)

 pmlindex = 0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsivz(:,:,pmlindex) = az(:,:,k)*dxizdz(:,:,k) + bzpml(:,:,k)*psivz(:,:,pmlindex)    
   dxizdz(:,:,k) = dxizdz(:,:,k) + psivz(:,:,pmlindex)
  endif 

 enddo

 div = dxixdx + dxiydy + dxizdz

 p = - c2rho0 * div - xi_z * gradp0_z
 rho = - rho0 * div - xi_z * gradrho0_z 

 call ddxyz(p, gradp_x, p, gradp_y, p, gradp_z)

 pmlindex = 0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsip(:,:,pmlindex) = az(:,:,k)*gradp_z(:,:,k) + bzpml(:,:,k)*psip(:,:,pmlindex)    
   gradp_z(:,:,k) =  gradp_z(:,:,k) + psip(:,:,pmlindex)
  endif 

 enddo


 RHSv_x = - rhoinv * gradp_x 

 RHSv_y = - rhoinv * gradp_y 

 RHSv_z = - rhoinv * gradp_z 

 do k= 1,nz
   RHSv_z(:,:,k) = RHSv_z(:,:,k) - rho(:,:,k)*g(k)*rhoinv(:,:,k)
   if (ABS(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) +  forcing*source_dep(k)
 enddo

 RHSxi_x = v_x
 RHSxi_y = v_y
 RHSxi_z = v_z


 do k=1,6
  scr(:,:,nz,k) = 0.0
  scr(:,:,1,k) = 0.0
 enddo



 if (DAMP_WAVES) call damp_velocity

 END SUBROUTINE MP_QUIET_PML_3D_DISPL

!================================================================================================


SUBROUTINE MP_MHD_PML_3D_DISPL

 ! CONSERVATIVE FORM OF THE FLUXES.

 ! I can't really do the conservative form - doesn't work for the incredibly small density 
 ! and forcing functions anyway. I don't know what the error is anyway
 implicit none
 integer i,j,k, pmlindex
 real*8, dimension(nx, dim2(rank), nz) :: flux1, flux2, flux3

 flux1 = - (xi_z * boy - xi_y * boz)
 flux2 = - (xi_x * boz - xi_z * box)
 flux3 = - (xi_y * box - xi_x * boy)

 call ddxyz(xi_x, dxixdx, xi_y, dxiydy, xi_z, dxizdz)
 pmlindex = 0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsivz(:,:,pmlindex) = az(:,:,k)*dxizdz(:,:,k) + bzpml(:,:,k)*psivz(:,:,pmlindex)    
   dxizdz(:,:,k) = dxizdz(:,:,k) + psivz(:,:,pmlindex)
  endif 

 enddo

 div = dxixdx + dxiydy + dxizdz

 p = - c2rho0 * div - xi_z * gradp0_z - xi_x * gradp0_x - xi_y * gradp0_y
 rho = - rho0 * div - xi_z * gradrho0_z - xi_x * gradrho0_x - xi_y * gradrho0_y

 call ddxyz(p, gradp_x, p, gradp_y, p, gradp_z)

 pmlindex = 0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsip(:,:,pmlindex) = az(:,:,k)*gradp_z(:,:,k) + bzpml(:,:,k)*psip(:,:,pmlindex)    
   gradp_z(:,:,k) =  gradp_z(:,:,k) + psip(:,:,pmlindex)
  endif 

 enddo

 call ddxyz(flux3, by, flux3, bx, flux2, dzflux2)
 call ddxyz(flux2, bz, flux1, flux3, flux1, dzflux1)


 ! NOTE THAT flux3 is overwritten in the third ddxyz call

 ! THE PML STUFF HAS TO COME HERE UNFORTUNATELY

 pmlindex=0
 do k=1,nz

  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
   pmlindex = pmlindex + 1

   RHSpsiinductionbx(:,:,pmlindex) = az(:,:,k)*dzflux2(:,:,k) &
			+ bzpml(:,:,k)*psiinductionbx(:,:,pmlindex)    

   dzflux2(:,:,k) = dzflux2(:,:,k) + psiinductionbx(:,:,pmlindex)


   RHSpsiinductionby(:,:,pmlindex) = az(:,:,k)*dzflux1(:,:,k) &
			+ bzpml(:,:,k)*psiinductionby(:,:,pmlindex)    

   dzflux1(:,:,k) = dzflux1(:,:,k) + psiinductionby(:,:,pmlindex)

  endif 
 
 enddo

 bx =   bx - dzflux2
 by = - by + dzflux1
 bz =   bz - flux3

 call ddxyz(by, curlbz, bz, curlbx,  bx, dzbx)
 call ddxyz(bz, curlby, bx, flux1, by, dzby)

 ! NOTE THAT FLUX1 IS OVERWRITTEN IN THE FINAL CALL

 ! THE PML STUFF HAS TO COME HERE UNFORTUNATELY

 pmlindex=0
 do k=1,nz
  if ((k .le. npmlbot) .or. (k .ge. (nz-npmltop+1))) then
    RHSpsidzbx(:,:,pmlindex) = az(:,:,k)*dzbx(:,:,k) &
			+ bzpml(:,:,k)*psidzbx(:,:,pmlindex)    

    dzbx(:,:,k) = dzbx(:,:,k) + psidzbx(:,:,pmlindex)

    RHSpsidzby(:,:,pmlindex) = az(:,:,k)*dzby(:,:,k) &
			+ bzpml(:,:,k)*psidzby(:,:,pmlindex)    

    dzby(:,:,k) = dzby(:,:,k) + psidzby(:,:,pmlindex)
   endif
 enddo


 curlbx = curlbx - dzby
 curlby = -curlby + dzbx
 curlbz = curlbz - flux1

 call cross(curlbx, curlby, curlbz, box, boy, boz, flux1, flux2, flux3)

 RHSv_x = rhoinv * (- gradp_x + flux1)
 RHSv_y = rhoinv * (- gradp_y + flux2)
 RHSv_z = rhoinv * (- gradp_z + flux3)


 call cross(curlbox, curlboy, curlboz, bx, by, bz, flux1, flux2, flux3)


 RHSv_x = RHSv_x + rhoinv*flux1 !- spongexyz * v_x
 RHSv_y = RHSv_y + rhoinv*flux2 !- spongexyz * v_y
 RHSv_z = RHSv_z + rhoinv*flux3 !- spongexyz * v_z


 do k= 1,nz
   RHSv_z(:,:,k) = RHSv_z(:,:,k) - rhoinv(:,:,k) * rho(:,:,k)*g(k)

   if (abs(k-e_rad) .LE. 4) &
    RHSv_z(:,:,k) = RHSv_z(:,:,k) + forcing*source_dep(k) 

 enddo 

 RHSxi_x = v_x
 RHSxi_y = v_y
 RHSxi_z = v_z

 do k=1,6
  scr(:,:,nz,k) = 0.
  scr(:,:,1,k) = 0.
 enddo


 if (DAMP_WAVES) call damp_velocity

 END SUBROUTINE MP_MHD_PML_3D_DISPL


!================================================================================================

END MODULE DISPLACEMENT
