! --------------------------------------------------------------------------
! 'params.i' : SIMULATION PARAMETERS
! --------------------------------------------------------------------------
! MPI Version of the Cartesian Solar Wave Simulator. 
! Copyright 2008, Shravan Hanasoge
!  
! W. W. Hansen Experimental Physics Laboratory
! Stanford University, Stanford
! CA 94305, USA
! Email: shravan@stanford.edu 
!
! Also see: http://soi.stanford.edu/
!									    
! --------------------------------------------------------------------------
!


! PARAMETERS OF THE SIMULATION
! nx, ny, nz are the numbers of grid points in the x,y,z directions resply.
! nz = 300 generally works well.
! xlength and ylength are the horizontal sides of the box, expressed in cgs
! A timestep of 2 seconds is generally pretty solid. Too large and the sim
! will explode. Too small and the expense is large.
integer nx, ny, nz
parameter (nx = 100, ny = 100, nz = 300)
real*8 xlength, ylength, timestep
parameter (xlength = 145.77 * 10**(8), ylength = xlength, timestep = 2.0)


! DIRECTORY INFORMATION 
!
! ENTER THE LOCATION OF THE BACKGROUND MODEL (i.e. the quiet Sun model)
character (LEN = *), PARAMETER :: file_data = '/data/seismo2/moradi/test_model_s'

! DIRECTORY FOR OUTPUT/ SAVED RESTART STATE (ASSUMING THEY ARE THE SAME)
character (LEN = *), PARAMETER :: directory = '/data/seismo1/hanasoge/rc_mhd_full/'

! THE FORCING FUNCTION
character (LEN = *), PARAMETER :: forcingfunc = '/data/seismo2/moradi/forcing100_145.fits'

! CADENCE OF THE FORCING FUNCTION (i.e. SAMPLING RATE)
real*8 cadforcing
parameter( cadforcing = 60.0)

! NOTE THIS fUNCTION STILL NEEDS TO BE IMPLEMENTED
! I HAVE THE BASIC MACHINERY IN PLACE BUT NEED DAMPING INFO
! DAMPING FUNCTION - IN ORDER TO DAMP WAVES - IN REAL SPACE (x,y,z)
logical, parameter :: DAMP_WAVES = .truE.


! TYPE OF SIMULATION (DEFAULT IS QUIET OR SOURCE/SOUND-SPEED ANOMALIES):: FLOWS OR MAGNETIC FIELDS 
logical, parameter :: magnetic = .truE.
logical, parameter :: FLOWS = .falsE.
logical, parameter :: TEST_IN_2D = .falsE.

! TYPE OF EQUATION
! IF DISPLACEMENT, SET DISPL = TRUE
logical, parameter :: DISPL = .falsE.

! VERTICAL BOUNDARY
! IF USE_PML IS FALSE, A SPONGE WILL BE IMPLEMENTED
logical USE_PML
parameter(USE_PML = .true.)


! ENTER THE LOCATION OF THE 3D pressure, density, and magnetic background files
! DEFAULT LOCATION IS THE SAME AS OUTPUT DIRECTORY
character (LEN = *), PARAMETER :: dirbackmag = '/data/seismo1/moradi/slim_test/ss_unstable/'

! --------

! STARTING TIME STEP; if starting off, time0 =0
! Otherwise, enter the index of the latest "full" file
integer, parameter :: time0 = 0

! HORIZONTAL BOUNDARY CONDITIONS
! Note since there are two ways to evaluate periodic horizontal
! boundary conditions, these two flags have to be set. If not FFTs,
! compact finite differences will be used to compute the periodic
! derivatives. Of course if periodic is set to false, then normal
! absorbing boundary conditions with derivatives computed using the
! compact finite differences are implemented.
logical, parameter :: PERIODIC = .TRUE.
logical, parameter :: USE_FFT = .FALSE.

! VERTICAL DERIVATIVE: USING EITHER SIXTH-ORDER ACCURATE COMPACT
! FINITE DIFFERENCES OR NINTH-ORDER ACCURATE OPTIMIZED EXPLICIT
! DIFFERENCES - SET COMPACT_FINITE_DIFF = .FALSE. IF YOU WANT TO
! USE THE LATTER
! RECOMMENDED: USE COMPACT FINITE DIFF.
logical, parameter :: compact_finite_diff = .TRUE.

! IF USE_PML = .TRUE., THEN SET THE FOLLOWING (DEFAULT VALUES WORK OKAY)
integer npmltop, npmlbot,nzpml
parameter(npmlbot = 7, npmltop = 10, nzpml = (npmlbot + npmltop))


! THE TOTAL WALL-TIME-LENGTH OF THE SIMULATION (IN HOURS)
real *8 wall_time
parameter(wall_time = 96.0)

! SOLAR TIME BEING SIMULATED (in hours)
real*8 solartime
parameter( solartime = 12.0)

! OBSERVATION HEIGHT RELATIVE TO PHOTOSPHERE (in cgs units)
! Generally 200 km above photosphere works well
! The sign indicates whether you wish to extract data above (+)
! or below (-) the photosphere
real*8 obsheight
parameter (obsheight = 20000000.0)

! CADENCE OF SLICE OUTPUT (in seconds; 30 s recommended)
real *8 outputcad
parameter (outputcad = 30.0)

! NUMBER OF ZEROES IN THE OUTPUT FILENAME
integer TIMESTAMP_SIZE 
parameter (timestamp_size = 6)

! WAVE EXCIATION DEPTH RELATIVE TO PHOTOSPHERE
! Generally 150 km below the photosphere is a good place to excite the waves
! excitdep is in cgs units
real*8 excitdep
parameter( excitdep = - 15000000.0)

